/*********************************************************************
 * 
 * Filename:      id3_frame_volume.c
 * Description:   Code for handling ID3 relative volume adjust frames.
 * Author:        Chris Vaill <cvaill@cs.columbia.edu>
 * Created at:    Thu Sep 26 23:57:00 2002
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 ********************************************************************/
#include <stdio.h>
#include <string.h>

#include "id3.h"


/* Assume buf points to the beginning of one channel's data in an RVA2
 * frame.  Return the encoded gain, in decibels. */
static float decode_rva2_gain(const char *buf)
{
    int gain_fp; /* fixed-point */

    gain_fp = *(signed char *)(buf + 1) << 8; /* first byte of gain */
    gain_fp |= *(unsigned char *)(buf + 2);   /* second byte of gain */

    return gain_fp / 512.0;
}


/* Assume buf points to the beginning of one channel's peak data in an
 * RVA2 frame (i.e. the fourth byte in the channel's data).  Return
 * the encoded peak, shifted to be a 32-bit value. */
static guint32 decode_rva2_peak(const unsigned char *buf)
{
    unsigned int peakbits = (unsigned int)*buf++;
    int shift, peakbytes;
    guint32 peak = 0;

    /* fast path for the common case */
    if (peakbits == 16)
	return (buf[0] << 24) + (buf[1] << 16);

    /*
     * We only care about the high 32 bits, so we care about the first
     * 5 bytes, because unfortunately, sizes that are not whole bytes
     * are padded on the msb side.
     *
     * e.g. a 33 bit peak has its first 7 bits zeroed, followed by 1
     * bit and four bytes of peak data.  We take the first four bytes
     * and shift left by 7, giving us 25 bits total.  We then take the
     * top 7 bits of the fifth byte to get our full 32 bits.
     */
    peakbytes = (peakbits + 7) >> 3;
    if (peakbytes > 4)
	peakbytes = 4;
    shift = ((8 - (peakbits & 7)) & 7) + (4 - peakbytes) * 8;

    /* get up to the first 4 bytes of the peak... */
    for (; peakbytes; peakbytes--) {
	peak <<= 8;
	peak += (unsigned int)*buf++;
    }

    /* ...and shift, so the value is out of 32 bits... */
    peak <<= shift;

    /* ...and finally, add part of the fifth byte, if necessary */
    if (peakbits > 32)
	peak += (unsigned int)*buf >> (8 - shift);

    return peak;
}


/*
 * Function id3_get_rva2_gain (frame)
 *
 *    Get the gain and peak info, that is coded in the RVA2 frame, and
 *    return it in the gn struct.  Return 0, or -1 on error.
 */
int id3_get_rva2_gain(id3_frame_t *frame, struct id3_gain *gn)
{
    const unsigned char *data;
    float gain;
    int i, chan, peakbytes;
    guint32 peak;

    /* Type check */
    if (frame->fr_desc->fd_id != ID3_RVA2)
	return -1;

    /* Check if frame is compressed */
    if (id3_decompress_frame(frame) == -1)
	return -1;

    /* init gains and peaks to 0 */
    memset(gn, 0, sizeof(struct id3_gain));

    data = (char *)frame->fr_data;

    /* skip past identification string */
    for (i = 0; i < frame->fr_size; i++)
	if (data[i] == '\0')
	    break;
    if (i >= frame->fr_size)
	return -1; /* ident string not terminated; bad frame data */

    /* cycle through channel fields */
    i++;
    while (i + 3 < frame->fr_size) {

	chan = data[i];

	/* decode the gain for this channel */
	gain = decode_rva2_gain(data + i);
	i += 3;

	/* decode the peak for this channel */
	peakbytes = (data[i] + 7) / 8;
	if (i + peakbytes >= frame->fr_size)
	    break;
	peak = decode_rva2_peak(data + i);
	i += 1 + peakbytes;

	if (chan <= 8) {
	    gn->channel[chan].ch_gain = gain;
	    gn->channel[chan].ch_peak = peak;
	}
    }

    return 0;
}
