#include "mpg123.h"
#include <string.h>
#include <stdlib.h>


#define ROUND(x) ((int)floor((x) + 0.5))

extern gchar *mpg123_filename;

float mpg123_current_gain = 0.0;

/* lookup table for fast gain adjustments */
static gint32 *volume_lut = NULL;
static struct id3_gain gain_info;
static int adjusting = FALSE;

/* track details of the lookup table, so we only rebuild if we need to */
static struct {
	AFormat lut_fmt;
	float   lut_gain;
	int     lut_limit;
	int     lut_boost;
} lut_data;

/*
 * A buffer of 8192 random numbers makes the period of the noise about
 * 85 ms, or 11.7 Hz, for 48 kHz stereo.  This should be low enough to
 * be unrecognizable except as noise.
 */
#define NOISEBUF_SZ 8192
static char *noisebuf;
static int noisebufi = 0;

/*
 * Returns 8 random bits, ie a number from -128 to +127.
 * Used for dithering.
 */
static int
rand8(void)
{
  int retval = (int)noisebuf[noisebufi];
  noisebufi = (noisebufi + 1) % NOISEBUF_SZ;
  return retval;
}


/*
 * Limiter function:
 *
 *        / tanh((x + lev) / (1-lev)) * (1-lev) - lev        (for x < -lev)
 *        |
 *   x' = | x                                                (for |x| <= lev)
 *        |
 *        \ tanh((x - lev) / (1-lev)) * (1-lev) + lev        (for x > lev)
 *
 * With limiter level = 0, this is equivalent to a tanh() function;
 * with limiter level = 1, this is equivalent to clipping.
 */
#define lmtr_lvl 0.5
static double
limiter(double x)
{
	double xp;

	if (x < -lmtr_lvl)
		xp = tanh((x + lmtr_lvl) / (1-lmtr_lvl)) * (1-lmtr_lvl) - lmtr_lvl;
	else if (x <= lmtr_lvl)
		xp = x;
	else
		xp = tanh((x - lmtr_lvl) / (1-lmtr_lvl)) * (1-lmtr_lvl) + lmtr_lvl;

	return xp;
}


void mpg123_voladjust_cleanup(void)
{
	adjusting = FALSE;
	if (noisebuf)
		g_free(noisebuf);
	noisebuf = NULL;
	if (volume_lut)
		g_free(volume_lut - 0x8000);
	volume_lut = NULL;
	return;
}


static void update_dither(void)
{
	int i;

	if (mpg123_cfg.use_rva2 && mpg123_cfg.enable_dither)
	{
		/* fill the noise buffer, which we will cycle through
		   when dithering */
		if (noisebuf == NULL) {
			noisebuf = g_malloc(NOISEBUF_SZ);
			for (i = 0; i < (NOISEBUF_SZ / sizeof(int)); i++)
				((int *)noisebuf)[i] = rand();
			noisebufi = 0;
			/* Note: noise shaping on the dithering noise
			   is not implemented, but this would be the
			   place to do it */
		}
	}
	else
	{
		if (noisebuf)
			g_free(noisebuf);
		noisebuf = NULL;
	}
}


static int gain_from_id3_tag(char *filename, struct id3_gain *gn)
{
	FILE *file;
	id3_t *id3d;
	id3_frame_t *id3frm;

	file = fopen(filename, "rb");
	if (file == NULL)
		goto err_out;

	id3d = id3_open_fp(file, 0);
	if (id3d == NULL)
		goto err_closefile;

	id3frm = id3_get_frame(id3d, ID3_RVA2, 1);
	if (id3frm == NULL)
		goto err_closeid3;

	if (id3_get_rva2_gain(id3frm, gn) == -1)
		goto err_closeid3;

	id3_close(id3d);
	fclose(file);
	return 1;

 err_closeid3:
	id3_close(id3d);
 err_closefile:
	fclose(file);
 err_out:
	memset(gn, 0, sizeof(struct id3_gain));
	return 0;
}


void mpg123_voladjust_update(char *filename)
{
	struct id3_gain newgain;
	float gain_db, gain_mult, scale;
	int i, use_boost, use_limiter = TRUE, need_update = FALSE;
	int samplemin, samplemax, center;
	int lut_samplemin, lut_samplemax, lut_center;
	AFormat fmt;
	guint32 peak;

	if (!mpg123_cfg.use_rva2)
	{
		mpg123_voladjust_cleanup();
		return;
	}

	update_dither();

	if (filename == NULL)
		filename = mpg123_filename;

	if (filename == NULL)
		return;

	if (!gain_from_id3_tag(filename, &newgain))
	{
		adjusting = FALSE;
		mpg123_current_gain = 0.0;
		memset(&gain_info, 0, sizeof(struct id3_gain));
		return;
	}

	fmt = mpg123_cfg.resolution == 16 ? FMT_S16_NE : FMT_U8;
	peak = newgain.channel[ID3_CHANNEL_MASTER].ch_peak;
	gain_db = newgain.channel[ID3_CHANNEL_MASTER].ch_gain;

	if (volume_lut == NULL)
	{
		/* the lookup table is 256k, but is deallocated if
		   volume adjustment is disabled */
		volume_lut = g_malloc(0x40000);
		/* move the pointer to the center of the allocated
		   region, so that we can use negative indices */
		volume_lut += 0x8000;
		need_update = TRUE;
	}

	gain_mult = pow(10, gain_db / 20.0);
	use_boost = mpg123_cfg.use_boost;
	if (use_boost)
		gain_mult *= 2;

	if (peak)
	{
		/*
		 * Figure out if the peak, multiplied by the gain,
		 * would clip.
		 *
		 * Some trickery here: we don't want to multiply our
		 * 32-bit peak by the gain, since it may overflow.
		 * Instead, we round it off to 24 bits. We then
		 * compare with 0xFFFF80, instead of 0xFFFFFF, to
		 * leave room for dithering.
		 */
		peak = ((peak >> 7) + 1) >> 1;
		if (peak * gain_mult <= 0xFFFF80)
			use_limiter = FALSE;

		if (use_limiter && mpg123_cfg.anticlip_mode == 1)
		{
			/* Reduce gain until the peak won't clip. */
			gain_mult = 0xFFFF80 / (float)peak;
			use_limiter = FALSE;
		}
	}

	if (!need_update
	    && lut_data.lut_fmt == fmt
	    && lut_data.lut_limit == use_limiter
	    && lut_data.lut_boost == use_boost
	    && fabs(lut_data.lut_gain - gain_db) < 0.0001)
	{
		/* the existing lookup table we have will work, no
		   need to build a new one */
		adjusting = TRUE;
		return;
	}

	/* We store 8 extra bits of precision in the table, for use in
	 * dithering. We can make this happen by simply multiplying
	 * the gain by 256. */
	gain_mult *= 256.0;

	if (fmt == FMT_U8)
	{
		samplemax = 0xFF;
		samplemin = 0;
		lut_samplemax = 0xFFFF;
		lut_samplemin = 0;
	}
	else
	{ /* fmt == FMT_S16_NE */
		samplemax = 0x7FFF;
		samplemin = ~samplemax;
		lut_samplemax = 0x7FFFFF;
		lut_samplemin = ~lut_samplemax;
	}

	/* leave space for dither, which ranges from -128 to +127 */
	lut_samplemax -= 127;
	lut_samplemin += 128;

	center = (samplemax + samplemin + 1) >> 1;
	lut_center = (lut_samplemax + lut_samplemin + 1) >> 1;

	/*
	 * Build the lookup table
	 */
	if (use_limiter)
	{
		/* apply gain, and use limiter to avoid clipping */
		scale = lut_center - lut_samplemin;
		for (i = samplemin; i < 0; i++)
			volume_lut[i] = lut_center + ROUND(scale * limiter((i-center) * gain_mult / scale));
		scale = lut_samplemax - lut_center;
		for (; i <= samplemax; i++)
			volume_lut[i] = lut_center + ROUND(scale * limiter((i-center) * gain_mult / scale));
	}
	else
	{
		/* just apply gain if it wouldn't clip anyway */
		for (i = samplemin; i <= samplemax; i++)
			volume_lut[i] = lut_center + ROUND((i-center) * gain_mult);
	}

	lut_data.lut_fmt = fmt;
	lut_data.lut_gain = gain_db;
	lut_data.lut_limit = use_limiter;
	lut_data.lut_boost = use_boost;
	mpg123_current_gain = gain_db;
	adjusting = TRUE;

	return;
}


void mpg123_voladjust(void *ptr, int length)
{
	AFormat fmt;
	int i;
	long sample;
	gint16 *ptr16 = NULL;
	guint8 *ptru8 = NULL;

	if (!adjusting || !mpg123_cfg.use_rva2 || volume_lut == NULL)
		return;

	fmt = mpg123_cfg.resolution == 16 ? FMT_S16_NE : FMT_U8;

	if (fmt == FMT_U8)
	{
		ptru8 = (guint8 *)ptr;
		if (mpg123_cfg.enable_dither)
		{
			for (i = 0; i < length; i++)
			{
				sample = volume_lut[*ptru8];
				sample += rand8();
				sample >>= 8;
				*ptru8 = sample;
				ptru8++;
			}
		}
		else
		{
			for (i = 0; i < length; i++)
			{
				*ptru8 = volume_lut[*ptru8] >> 8;
				ptru8++;
			}
		}
	}
	else
	{ /* fmt == FMT_S16_NE */
		ptr16 = (gint16 *)ptr;
		if (mpg123_cfg.enable_dither)
		{
			for (i = 0; i < length; i += 2)
			{
				sample = volume_lut[*ptr16];
				sample += rand8();
				sample >>= 8;
				*ptr16 = sample;
				ptr16++;
			}
		}
		else
		{
			for (i = 0; i < length; i += 2)
			{
				*ptr16 = volume_lut[*ptr16] >> 8;
				ptr16++;
			}
		}
	}
}
